import paramiko
import os
import stat
import re
from UtilsRPE.Functions import load_vault
import pickle


def remote_isdir(attr):
    return stat.S_ISDIR(attr.st_mode)


class SSH:
    def __init__(self, user, host, remote_scratch=""):
        # Init
        self.user = user
        self.host = host
        self._host_config_id = None
        self._ssh = None
        self._ssh_config = None
        self._user_config_file = None
        self._host_config = None
        self.sftp = None
        self.remote_scratch = remote_scratch

        # SSH
        self.connect()
        self.init_transport()

    def init_transport(self):
        self.sftp = self._ssh.open_sftp()

    def get(self, remote_path, local_path):
        if not self.sftp:
            self.init_transport()
        self.sftp.get(remote_path, local_path)

    def put(self, local_path, remote_path):
        if not self.sftp:
            self.init_transport()
        self.sftp.put(local_path, remote_path)

    def connect(self):
            """
            Creates ssh connection to host

            :return: True if connection is created, False otherwise
            :rtype: bool
            """
            try:
                self._ssh = paramiko.SSHClient()
                self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._ssh_config = paramiko.SSHConfig()
                self._user_config_file = os.path.expanduser("~/.ssh/config")
                if os.path.exists(self._user_config_file):
                    with open(self._user_config_file) as f:
                        # noinspection PyTypeChecker
                        self._ssh_config.parse(f)
                self._host_config = self._ssh_config.lookup(self.host)
                if 'identityfile' in self._host_config:
                    self._host_config_id = self._host_config['identityfile']
                if 'proxycommand' in self._host_config:
                    self._proxy = paramiko.ProxyCommand(self._host_config['proxycommand'])
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id, sock=self._proxy)
                else:
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id)
                return True
            except IOError as e:
                print('Can not create ssh connection to {0}: {1}', self.host, e.strerror)
                return False

    def list_experiments_(self, expid):
        if not self.sftp:
            self.init_transport()
        experiment_path = "/gpfs/scratch/pr1egi00/pr1egi23/%s" % expid

        list_of_experiments = []

        # Enter in experiment folders
        experiment_pattern = "[0-9]+"
        member_pattern = "[a-z][0-9]+"
        variable_pattern = "ref_[0-9]+"

        for attr in list(self.sftp.listdir_attr(experiment_path)):
            filename = attr.filename
            if re.match(experiment_pattern, filename):
                if remote_isdir(attr):
                    member_path = "%s/%s" % (experiment_path, filename)
                    member_list = list(self.sftp.listdir_attr(member_path))
                    for member in member_list:
                        if re.match(member_pattern, member.filename):
                            if remote_isdir(member):
                                member_filepath = "%s/%s" % (member_path, member.filename)
                                list_of_experiments.append(member_filepath)
                                variable_list = list(self.sftp.listdir_attr(member_filepath))
                                for variable in variable_list:
                                    if remote_isdir(variable):
                                        if re.match(variable_pattern, variable.filename):
                                            variable_filepath = "%s/%s" % (member_filepath, variable.filename)
                                            list_of_experiments.append(variable_filepath)

        return list_of_experiments

    def find_results(self, expid):
        namelist_pattern = "emulator_variable_precisions\(([0-9]+)\) *= *([0-9]+)"
        experiments = self.list_experiments_(expid)
        database = {"error": {}, "precisions": {}}
        for experiment in experiments:
            accuracy_filepath = "%s/accuracy.txt" % experiment
            namelist_filepath = "%s/namelist_precisions" % experiment
            with self.sftp.open(accuracy_filepath) as f:
                error = float(f.read())
            database["error"][experiment] = error
            database["precisions"][experiment] = {}
            with self.sftp.open(namelist_filepath) as f:
                for l in f:
                    m = re.match(namelist_pattern, l)
                    if m:
                        var_id, precision = [ int(x) for x in m.groups(0)]
                        database["precisions"][experiment][var_id] = precision

        return database

    def execute(self, command):
        return self._ssh.exec_command(command)

    def check_completed_experiments(self):
        if not self.sftp:
            self.init_transport()
        experiment_path = self.remote_scratch

        list_of_experiments = []

        # Enter in experiment folders
        experiment_pattern = "[0-9]+"
        member_pattern = "[a-z][0-9]+"
        variable_pattern = "ref_[0-9]+"

        for attr in list(self.sftp.listdir_attr(experiment_path)):
            filename = attr.filename
            if re.match(experiment_pattern, filename):
                if remote_isdir(attr):
                    member_path = "%s/%s" % (experiment_path, filename)
                    member_list = list(self.sftp.listdir_attr(member_path))
                    for member in member_list:
                        if re.match(member_pattern, member.filename):
                            if remote_isdir(member):
                                member_filepath = "%s/%s" % (member_path, member.filename)
                                list_of_experiments.append(member_filepath)
                                variable_list = list(self.sftp.listdir_attr(member_filepath))
                                for variable in variable_list:
                                    if remote_isdir(variable):
                                        if re.match(variable_pattern, variable.filename):
                                            variable_filepath = "%s/%s" % (member_filepath, variable.filename)
                                            list_of_experiments.append(variable_filepath)

        return list_of_experiments

if __name__ == "__main__":
    # Load vault
    vault = load_vault("/home/Earth/otintopr/RPE-ROMS/AutoRPE/vault.pkl")

    # SSH to remote machine and download results
    hostname = "mn1.bsc.es"
    user = "pr1egi23"
    expid = "a1bo"

    s = SSH(user=user, host=hostname)
    results = s.find_results(expid)

    variables = [variable for variable in vault.variables if variable.id is not None]
    var_dict = {}

    for variable in variables:
        var_dict[variable.id] = variable

    for variable in variables:
        variable.error = 1000

    for key, error in results["error"].items():
        for var_id, precision in results["precisions"][key].items():
            if precision == 10:
                if error < var_dict[var_id].error:
                    var_dict[var_id].error = error

    ordered_vars = var_dict.values()

    ordered_vars = sorted(ordered_vars, key=lambda x: -x.error)

    for var in ordered_vars:
        print("%4i\t\t%8.2e" % (var.id, var.error))

    save_vault = True
    if save_vault:
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        new_vault_name = "vault_with_processed_a1bo.pkl"
        pickle.dump(vault, open(new_vault_name, "wb"))
