

def completed_jobs_have_successful_defined(all_jobs):
    # Check if completed jobs have a defined is_successful property
    for job in all_jobs:
        if job.is_completed:
            if job.is_successful is None:
                raise AssertionError("Completed but without a is_successful defined!")


def successful_job_descendants(all_jobs):
    # A successful job should not have a non-banned descendant that is not successful
    for job in all_jobs:
        if job.is_completed:
            if job.is_successful is True:
                for descendant in job.descendants():
                    if descendant.is_banned is False:
                        if descendant.is_successful is not True:
                            raise AssertionError("Descendant of a successful job unsuccessful")


def non_successful_job_ancestors(all_jobs):
    # An ancestor of a non successful job should not be successful
    for job in all_jobs:
        # For any given completed job that is not banned
        if job.is_completed and job.is_banned is False and job.is_successful is False:
            # Look at all the ancestors to assert that any of them is successful
            for ancestor in job.ancestors():
                if ancestor.is_banned is False and ancestor.is_successful is True:
                    raise AssertionError("Ancestor of a unsuccessful job successful")


def check_duplicates(queue):
    # Check for duplicates
    duplicated_completed_running = [job for job in queue.completed if job in queue.running]
    duplicated_completed_queue = [job for job in queue.completed if job in queue]
    duplicated_running_queue = [job for job in queue.running if job in queue]

    all_duplicated = duplicated_completed_running + duplicated_completed_queue + duplicated_running_queue

    duplicated_jobs = len(all_duplicated)

    if duplicated_jobs > 0:
        names = " ".join([j.identifier() for j in all_duplicated])
        raise AssertionError("Duplicated jobs!: %s" % names)


def check_banned_variables(all_jobs, ban_list):
    # Check that any non banned job has banned variables
    for job in all_jobs:
        if job.is_banned is False:
            for variable in ban_list:
                if variable in job.variable_set:
                    all_variables_banned = True
                    for v in job.variable_set:
                        if v not in ban_list:
                            all_variables_banned = False
                    if all_variables_banned is False:
                        raise AssertionError("Banned variable present in variable set!")


def job_not_present_in_any_list(all_jobs, queue):
    for job in all_jobs:
        anywhere = job in queue or job in queue.running or job in queue.completed
        if not anywhere:
            raise AssertionError("Job has been lost!")


def banned_job_has_descendants_running(all_jobs, queue):
    for job in all_jobs:
        if job.is_banned:
            for descendant in job.descendants():
                if descendant in queue.running:
                    raise AssertionError("Descendant of a banned set running.")


def check_job_in_running_is_successful(all_jobs, queue):
    for job in all_jobs:
        if job.is_successful:
            if job in queue.running:
                raise AssertionError("Job in running with is_successful defined")


def jobs_in_queue_have_same_hash(all_jobs, queue):
    jobs = [job for job in all_jobs if job in queue or job in queue.running]

    hashes = []
    for job in jobs:
        if job.hash in hashes:
            raise AssertionError("Two jobs with same hash in queues")
        hashes.append(job.hash)


def jobs_in_queue_with_a_non_banned_unsuccessful_descendant(all_jobs, queue):
    jobs = [job for job in all_jobs if job in queue or job in queue.running]

    for job in jobs:
        if job in queue.running:
            if job.unsuccessful_descendant:
                raise AssertionError("Job running with an unsuccessful non-banned descendant!")


def unsuccessful_unbanned_jobs_with_a_single_variable(all_jobs, queue):
    for job in all_jobs:
        if not job.child:
            if job.is_banned is False and job.is_successful is False:
                raise AssertionError("unsuccessful_unbanned_jobs_with_a_single_variable")


def no_jobs_left_in_queue(queue):
    if len(queue) == 0 and len(queue.running) == 0:
        raise EmptyQueuesError("No jobs queuing and no running jobs!")


# Attempt to assert that nothing nasty happens
def check_consistency(seed_set, queue, ban_list):
    all_jobs = seed_set.descendants()
    all_jobs.append(seed_set)

    job_not_present_in_any_list(all_jobs, queue)

    completed_jobs_have_successful_defined(all_jobs)

    successful_job_descendants(all_jobs)

    non_successful_job_ancestors(all_jobs)

    check_duplicates(queue)

    check_banned_variables(all_jobs, ban_list)

    banned_job_has_descendants_running(all_jobs, queue)

    check_job_in_running_is_successful(all_jobs, queue)

    jobs_in_queue_with_a_non_banned_unsuccessful_descendant(all_jobs, queue)

    unsuccessful_unbanned_jobs_with_a_single_variable(all_jobs, queue)
    # jobs_in_queue_have_same_hash(all_jobs, queue)
    no_jobs_left_in_queue(queue)


def checkpoint(queue):
    print("Saving checkpoint!", end="\r")
    import pickle

    all_jobs = queue.pending + queue.running + queue.completed

    # Save the communicator
    communicator = all_jobs[0].communicator
    for job in all_jobs:
        job.communicator = None
    # Dump object
    pickle.dump(queue, open("checkpoint.pkl", "wb"))

    # Restore the communicator
    for job in all_jobs:
        job.communicator = communicator
    print("Checkpoint saved!")


def recover_checkpoint(communicator, filename="checkpoint.pkl"):
    import pickle as pkl
    print("Loading checkpoint!", end="\r")
    queue = pkl.load(open(filename, "rb"))
    all_jobs = queue.pending + queue.running + queue.completed
    for job in all_jobs:
        job.communicator = communicator
    job_with_no_parent = [job for job in all_jobs if job.parent is None]
    grandfather = job_with_no_parent[0]
    ban_list = grandfather.banned_variables
    print("Checkpoint loaded!!")
    return queue, grandfather, ban_list


def saving_state(_queue, filename="state.pkl"):
    import pickle as pkl
    log = {"COMPLETED": {}, "RUNNING": [], "QUEUING": []}
    for _job in _queue.running:
        log["RUNNING"].append(_job.identifier())
    for _job in _queue:
        log["QUEUING"].append(_job.identifier())
    for _job in _queue.completed:
        log["COMPLETED"][_job.identifier()] = (_job.is_successful, _job.is_banned)
    with open(filename, "wb") as output_log:
        pkl.dump(log, output_log, protocol=2)


class EmptyQueuesError(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super(EmptyQueuesError, self).__init__(message)


