# Script to compute the normalized Root Mean Square Error between a reference simulation
# and a simulation to be evaluated.
# The script reads the following mandatory command line arguments:
# -r: reference file
# -f: file to evaluate

import numpy as np
from os.path import isfile
import pickle as pkl


# Hardcoded path (Warning!)
mask_filename = "/gpfs/scratch/pr1egi00/pr1egi23/NEMO-2Dmode/EnsembleOutput/mesh_mask.nc"


def read_arguments():
    # Using optparse library to get command line options and arguments
    from optparse import OptionParser

    # Reading the command line arguments and
    # Creating the parser
    parser = OptionParser()
    # Adding options
    # Option for generating simplified output
    parser.add_option("-r", help="Reference File", dest="reference", type="string", default="year_01")
    parser.add_option("-f", help="File to evaluate", dest="to_evaluate", type="string")

    # Parse arguments and options
    (options, args) = parser.parse_args()
    print(options)

    _evaluated_filename_pattern = options.to_evaluate
    print(_evaluated_filename_pattern)
    _reference_filename_pattern = options.reference
    return _reference_filename_pattern, _evaluated_filename_pattern


reference_filename_pattern, evaluated_filename_pattern = read_arguments()

# List of grid files created during NEMO simulations
grids = ["scalar", "grid_T", "grid_U", "grid_V", "grid_W"]

# Create file pattern for every grid type
for grid in grids:
    evaluated_filename_pattern = evaluated_filename_pattern.replace(grid, "%s")
    reference_filename_pattern = reference_filename_pattern.replace(grid, "%s")

# Define output file
output_file = evaluated_filename_pattern % "RMSE"
output_file = output_file.replace(".nc", ".pkl")


def load_file(_filename):
    # Routine that reads a netcdf file
    from netCDF4 import Dataset as NetCDFFile
    try:
        nc = NetCDFFile(_filename)
    except OSError:
        print("File %s not found!" % _filename)
        exit(1)
    return nc


def load_var(_filename, _var_name):
    # Reads a netCDF file and returns a specific variable converted to numpy array
    import numpy as np
    nc = load_file(_filename)
    # Load longitudes, latitudes and bathymetry data
    _data = nc.variables[_var_name]
    _data = np.array(_data)
    # To solve problem with _FillValue
    _data[_data == 1e20] = 0
    return _data


def mask_var(data, _mask):
    # Masks a variable using a mask
    number_of_variables = len(data.shape)
    if number_of_variables == 4:
        t, z, x, y = data.shape
    elif number_of_variables == 3:
        t, x, y = data.shape
        z = None
    elif number_of_variables == 2:
        t, x = data.shape
        y = z = None
    else:
        raise AssertionError("Case not considered")

    _mask = np.array([_mask for i in range(t)])
    if z is None:
        _mask = np.sum(_mask, axis=2)
    if y is None:
        _mask = np.sum(_mask, axis=1)
    _mask = np.logical_not(_mask)
    try:
        return np.ma.array(data, mask=_mask)
    except np.ma.core.MaskError:
        raise AssertionError(_mask.shape, data.shape)


class DataFile:
    # Class to load variables and keep it in memory
    def __init__(self, file_name):
        self.filename = file_name
        self.vars = {}

    def load_var(self, var_name, mask=None):
        if mask is None:
            self.vars[var_name] = load_var(self.filename, var_name)
        else:
            data = load_var(self.filename, var_name)
            number_of_variables = len(data.shape)
            if number_of_variables == 4:
                t, z, x, y = data.shape
            elif number_of_variables == 3:
                t, x, y = data.shape
                z = None
            elif number_of_variables == 2:
                t, x = data.shape
                y = z = None
            else:
                raise AssertionError("Case not considered")

            mask = np.array([mask for i in range(t)])
            if z is None:
                mask = np.sum(mask, axis=2)
            if y is None:
                mask = np.sum(mask, axis=1)
            mask = np.logical_not(mask)
            self.vars[var_name] = np.ma.array(data, mask=mask)

    def __getitem__(self, key):
        return self.vars[key]


def root_mean_squared_error(predictions, targets):
    # Routine to compute the RMSE between two arrays.

    # If there's less time-steps for the predictions than for the targets, take only the needed part of the targets
    if predictions.shape != targets.shape:
        targets = targets[:predictions.shape[0]]

    # Compute the differences
    try:
        differences = predictions - targets
    except AssertionError:
        # Return an assertion error if the operation failed.
        raise AssertionError

    # Compute the square of the difference
    differences_squared = differences ** 2

    # Compute the mean of the difference value squared
    if len(predictions.shape) == 4:
        mean_of_differences_squared = differences_squared.mean(axis=(3, 2, 1))  # the MEAN of ^
    elif len(predictions.shape) == 3:
        mean_of_differences_squared = differences_squared.mean(axis=(1, 2))  # the MEAN of ^

    else:
        raise AssertionError("Variable has wrong dimensions")

    # Compute the square root value of the mean
    rmse_val = np.sqrt(mean_of_differences_squared)

    return rmse_val


# Create a dictionary that will store the RMSE values for all the variables in all the grid types.
rmse_dict = {}

# Loop over grids
for grid in grids:

    # Create dictionary for the variables
    rmse_dict[grid] = {}

    # Set the proper reference filename for the current grid
    reference_filename = reference_filename_pattern % grid

    # Check its existence
    assert (isfile(reference_filename))

    # Load mask for non scalar grids
    if grid != "scalar":
        # Load Mask
        mask = DataFile(mask_filename)
        if grid == "grid_W":
            mask_name = "tmask"
        elif grid == "grid_T":
            mask_name = "tmask"
        elif grid == "grid_U":
            mask_name = "umask"
        elif grid == "grid_V":
            mask_name = "vmask"
        else:
            raise AssertionError("Unknown grid")
        mask.load_var(mask_name)
        mask = mask[mask_name]
    else:
        mask = None

    # Read list of variables from reference file
    reference_file = load_file(reference_filename)
    var_names = reference_file.variables.keys()
    del reference_file

    # Loop over list of variables in the file
    for var_name in var_names:
        ref_data = load_var(reference_filename, var_name)
        if mask is not None:
            try:
                ref_data = mask_var(ref_data, mask)
            except AssertionError:
                continue
        evaluated_filename = evaluated_filename_pattern % grid
        evaluated_data = load_var(evaluated_filename, var_name)
        try:
            evaluated_data_rmse = root_mean_squared_error(evaluated_data, ref_data)
            print("Var %s computed" % var_name)
            del ref_data
            del evaluated_data
        except AssertionError:
            continue
        rmse_dict[grid][var_name] = evaluated_data_rmse

pkl.dump(rmse_dict, open(output_file, "wb"))
