# Script that automatically creates different versions of a routine with different precisions

from AutoRPE.UtilsRPE.Cleaner import merge_lines
from AutoRPE.UtilsRPE.MainFunctions import parse_sources,  preprocess_sources
from AutoRPE.UtilsRPE.Functions import load_vault, remove_comments
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.CodeTools import replace_real_declaration
from os import listdir, mkdir
from os.path import join, isdir, isfile
from shutil import rmtree
import re
from math import factorial
import itertools

# Few functions
def merge_declaration_lines(code_text):
    lines = [ l for l in code_text.split("\n")]
    for _index, _line in enumerate(lines):
        line = remove_comments(_line).strip()
        if line.count("&") and line.count("::"):
            i = 1
            while line and line.strip()[-1] == "&":
                line1 = line
                line2 = remove_comments(lines[_index + i]).strip()
                line = line1.replace("&", "").rstrip() + line2
                i += 1
            line = line.replace("&", "")
            line = re.sub(", +", ", ", line)
            line = re.sub(":  +", ":  ", line)
            lines[_index] = line
            for j in range(1, i):
                lines[_index + j] = ""
    return "\n".join(lines).strip()

def replace_routine_name(old_name, new_name, code_text):
    pattern = r"(subroutine +\b)%s\b" % old_name
    pat = re.compile(pattern, re.I)
    replacement = r"\1 %s" % new_name
    return pat.sub(replacement, code_text)
 
    

def find_declaration_line(variable, routine_code):
    import re
    from AutoRPE.UtilsRPE.Functions import remove_comments
    # Declaration pattern
    # TODO: This might have issues with split lines
    pattern = r"::.*\b%s\b" % variable.name
    exception_pattern = r"&*\b%s\b" % variable.name
    real_pattern = r"real.*::"
    lines = [l for l in routine_code.split("\n")]

    current_block = CurrentBlock(vault[variable.routine.module.name].main())

    for _index, _line in enumerate(lines):
        _line = remove_comments(_line).strip()
        if re.search(pattern, _line):
            if current_block.routine == variable.routine:
                return _index
        else:
            current_block = current_block.in_which_block(_line, vault=vault)

    print("Declaration of variable %s with id %i not found in module %s" % (variable.name,variable.id, variable.routine.module.name))
    return None



def obtain_rotine_source(routine, sources_path):
    source_file = join(sources_path, "%s.f90" % routine.module.name)
    with open(source_file) as f:
       pattern = r"routine.*\b%s\b" % routine.name
       lines = [l for l in f]
       indices = []
       for index, line in enumerate(lines):
          line = remove_comments(line).strip()
          if re.search(pattern, line, re.I):
             indices.append(index)
       routine_code = "".join(lines[indices[0]:indices[1]+1])

    routine_code = merge_declaration_lines(routine_code)
    return routine_code

if __name__ == "__main__":
    # Getting Command line options
    from optparse import OptionParser
    help_text = """
    This script takes as arguments the input source (-s/--input-source)
    and the name of a routine (-r/--routine) and creates different versions
    of this routine accouting for all different precision combinations that can happen.
    An additional argument for the vault (-v/--vault) allows to skip the
    creation of the vault if this is provided.
    """
    parser = OptionParser(usage=help_text)
    parser.add_option("-v", "--vault", dest="vault", default=None, metavar="FILE")
    parser.add_option("-s", "--input-source", dest="input_source", default=None, metavar="FILE")
    parser.add_option("-r", "--routine", dest="routine_name", default=None)

    (options, args) = parser.parse_args()
    vault_path = options.vault
    path_to_input_sources = options.input_source
    routine_name = options.routine_name


    if path_to_input_sources is None:
       print("\nArgument -s/--input-source and -r/--routine are mandatory\n")
       parser.print_help()
       exit(1)

    if routine_name is None:
       print("\nArgument -r/--routine is mandatory\n")
       parser.print_help()
       exit(1)

    path_to_preprocessed_sources = "./tmp_processedsources/"
    # In case no vault has been provided it will create it from the sources in the input sources folder
    if vault_path is None:
        # Paths to source folders
        # Path_to_input_sources
        if not isdir(path_to_preprocessed_sources):
           mkdir(path_to_preprocessed_sources)
        # List of files that should not be processed, as red from "list_of_files_to_keep_unmodified.txt"
        preprocess_blacklist = listdir(path_to_input_sources)

        # Format text + replace real declarations with type(rpe_var)
        preprocess_sources(path_to_input_sources, path_to_preprocessed_sources, blacklist=preprocess_blacklist)

        # Obtain source information
        vault = parse_sources(path_to_preprocessed_sources, save_vault=True)
    else:
        vault = load_vault(vault_path)

    routine = None
    for procedure in vault.procedures:
       if procedure.name == routine_name:
          routine = procedure
          break
    if routine is None:
        print("Routine %s not found in vault." % routine_name)
        exit(0)
    
    
    routine_code = obtain_rotine_source(routine, path_to_input_sources)
    routine_lines = routine_code.split("\n")
    sensitive_variables = [v for v in routine.variables if v.is_dummy_argument and v.type == "real"]
    # All possible combinations of precisions
    precision_combinations = list(itertools.product(["dp","sp"], repeat=len(sensitive_variables)))

    code_versions = {}
    interface = "module procedure "
    for c_index, combination in enumerate(precision_combinations):
       version_lines = routine_lines[:]
       new_routine_name = "%s_%04i" % (routine.name, c_index)
       for index, variable in enumerate(sensitive_variables):
          variable_precision = combination[index]
          dec_line = find_declaration_line(variable, routine_code)
          original_line = version_lines[dec_line]
          new_line = replace_real_declaration(variable, original_line, variable_precision)
          version_lines[dec_line] = new_line
       version_text = "\n".join(version_lines).strip()
       version_text = replace_routine_name(routine.name, new_routine_name,version_text)
       code_versions[c_index] = version_text
       if c_index == 0:
          interface += "%s" % new_routine_name
       else:
          interface += ", %s" % new_routine_name

    interface = "interface %s\n           %s\nend interface %s" % (routine.name, interface, routine.name)
    print(interface)
    for key, code in code_versions.items():
       print(code)
    # Deleting temporary folders
    if isdir(path_to_preprocessed_sources):
       rmtree(path_to_preprocessed_sources)
