from os.path import isdir, abspath
from os import listdir, mkdir
from shutil import rmtree
import networkx as nx

import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
from AutoRPE.UtilsRPE.MainFunctions import parse_sources, preprocess_sources
from AutoRPE.UtilsRPE.Functions import load_vault
from AutoRPE.UtilsRPE.TrackDependency import track_dependencies, propagate_dependencies


def get_command_line_arguments():
    """
    Returns a list of files that have been provided as a command line argument
    :return: list of files
    """

    # Parse and assert command line options
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--vault", default=None,
                        help="name of the input vault pkl file, if any")
    parser.add_argument("-w", "--working-precision", default=23,
                        help="The precision used for real numbers in the parsed code")
    parser.add_argument("-o", "--output-vault", default="vault.pkl",
                        help="Name of the output vault file")
    parser.add_argument("-t", "--omit-tracking", default=True, action="store_false",
                        help="Whether or not you are going to track dependency")
    parser.add_argument("-g", "--export-graph", default=False, action="store_true",
                        help="Whether or not you want to store the created graph")
    parser.add_argument("-i", "--input-source", default=None,
                        help="Path to input fortran files")

    args = parser.parse_args()

    if args.input_source is None and args.vault_path is None:
        parser.error("One of the arguments -s/--input-source or -v/--vault is mandatory")

    return args


if __name__ == "__main__":

    args = get_command_line_arguments()

    vault_path = args.vault
    wp = args.working_precision
    output_vault = args.output_vault
    path_to_input_sources = args.input_source
    export_graph = args.export_graph
    track = args.omit_tracking

    # Set the proper working precision
    VariablePrecision.wp = VariablePrecision.precision2type[wp]
    print("Creating a database from sources found in:\n   %s\n\n" % path_to_input_sources)

    # In case no vault has been provided it will create it from the sources in the input sources folder
    if vault_path is None:
        # Paths to source folders
        # Path_to_input_sources
        path_to_preprocessed_sources = "./tmp_processedsources/"
        path_to_output_sources = "./tmp_outputsources/"
        if not isdir(path_to_preprocessed_sources):
            mkdir(path_to_preprocessed_sources)
        if not isdir(path_to_output_sources):
            mkdir(path_to_output_sources)
        # List of files that should not be processed, as red from "list_of_files_to_keep_unmodified.txt"
        preprocess_blacklist = listdir(path_to_input_sources)

        # Format text + replace real declarations with type(rpe_var)
        preprocess_sources(path_to_input_sources, path_to_preprocessed_sources, blacklist=preprocess_blacklist)

        # Obtain source information
        vault = parse_sources(path_to_preprocessed_sources, save_vault=False)
    else:
        vault = load_vault(vault_path)

    real_variables = [v for v in vault.variables if v.type in VariablePrecision.real_types]

    length = len(real_variables)

    # Assigning id and precision
    for index, var in enumerate(real_variables):
        var.id = index
        var.precision = VariablePrecision.type2precision[var.type]

    if track:
        # Track dependencies and propagate it
        # Create empty graph
        graph = nx.DiGraph()
        track_dependencies(path_to_preprocessed_sources, vault, graph=graph)
        if export_graph:
            nx.readwrite.gexf.write_gexf(graph, "./graph.gexf")
        propagate_dependencies(vault, graph=graph)

    # Save the vault with the variable precisions updated.
    # This workaround is needed to allow pickle to write the vault file
    import sys
    import pickle

    sys.setrecursionlimit(20000)
    pickle.dump(vault, open(output_vault, "wb"))
    print("Vault successfully dumped to ", abspath(output_vault))

    # Deleting temporary folders
    if isdir(path_to_preprocessed_sources):
        rmtree(path_to_preprocessed_sources)
    if isdir(path_to_output_sources):
        rmtree(path_to_output_sources)
