from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ReadSourceFile, ObtainSourcefileInfo
from AutoRPE.UtilsRPE.BasicStructures import Routine, Interface, Function
from AutoRPE.UtilsRPE.Counter import Counter
from AutoRPE.UtilsRPE.Splitter import split_semicolon, split_elements_in_operations, split_read_arguments, \
    split_line_container_and_arguments
from AutoRPE.UtilsRPE.Cleaner import remove_comments, remove_empty_lines
from AutoRPE.UtilsRPE.Getter import get_type_of_contents, get_interface_from_arguments, is_keyword_argument
from AutoRPE.UtilsRPE.NatureDeterminer import has_intrinsics, is_a_number, has_operations
from AutoRPE.UtilsRPE.BasicFunctions import remove_literals, close_brackets, literal_mask
from AutoRPE.UtilsRPE.CallManager import find_call_arguments,find_called_routine, calls_to_routines, \
    calls_to_functions, find_called_function
from AutoRPE.UtilsRPE.Inserter import replace_variable_exact_match, add_rpe_var_to_arguments, \
    add_rpe_to_argument, add_real_to_argument, add_val_to_argument
from AutoRPE.UtilsRPE.TemporalVariable import use_temporal_variable
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import AutoRPE.UtilsRPE.DerivedTypeDealer as DerivedTypeDealer
import AutoRPE.UtilsRPE.IntrinsicDealer as IntrinsicDealer
import AutoRPE.UtilsRPE.Error as Error

import re
import warnings


class ImplementRPE(ReadSourceFile):
    def __init__(self, _filename, vault):
        super(ImplementRPE, self).__init__(_filename)

        # Dictionaries and lists about source info
        self.source_info = vault
        if self.source_info:
            module_name = _filename.replace(".f90", "")
            if module_name.count("/"):
                module_name = module_name.split("/")[-1]
            self.module = self.source_info[module_name]
        else:
            self.module = None

    # Split lines which contains more than one command
    # TODO: We should be sure that we are not splitting by characters in literals
    def split_lines(self):

        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            if line.count(";"):
                parts = split_semicolon(line)
                self.lines[_index] = parts[0]
                for i in range(1, len(parts)):
                    self.lines.insert(_index + 1, parts[i])
                    length += 1
                    _index += 1
            _index += 1
        self.rebuild_text()

    # Split declarations of variables putting a single variable in each line
    def split_declarations(self):
        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            line = remove_literals(line)
            if line.count("::"):
                attributes, arguments = line.split("::")
                arguments = find_call_arguments("c(%s)" % arguments)
                if len(arguments) > 1:
                    new_line = "%s :: %s" % (attributes, arguments[0])
                    self.lines[_index] = new_line
                    for argument in arguments[1:]:
                        new_line = "%s :: %s" % (attributes, argument)
                        self.lines.insert(_index + 1, new_line)
                        length += 1
                        _index += 1

            _index += 1
        self.rebuild_text()

    # Remove comments from a file
    def remove_all_comments(self):
        for _index, _line in enumerate(self.lines):
            self.lines[_index] = remove_comments(_line)


    # If parameter declaration is made in Fortran 77 style
    def remove_F77_parameter_declaration(self):

        pattern = '^\s*PARAMETER\s*\(\s*[\w+=\s*]'

        for line_index, line in enumerate(self.lines):

            # Search for lines with the 'PARAMETER' statement
            match = re.match(pattern, line, re.I)
            if match:
                param_line = line.replace(")", "")

                # Store parameter names
                parameters_name = [parameter.split("=")[0].strip() for parameter in
                                   param_line.split("(")[1].split(",")]

                # Store parameter values
                parameters_val = [parameter.split("=")[1].strip() for parameter in
                                  param_line.split("(")[1].split(",")]

                for parameter_name, parameter_value in zip(parameters_name, parameters_val):

                    # Search for each parameter declaration backward in the file
                    for dec_line_index, dec_line in enumerate(reversed((self.lines[:line_index]))):
                        parameter_pattern = '.*::\s*' + parameter_name + '$'
                        param_declaration = re.match(parameter_pattern, dec_line, re.I)
                        if (line_index - dec_line_index - 1) == 0:
                            raise AssertionError("declaration of parameter %s not found in file %s" % (
                                parameter_name, self.filename))
                        # Put back in place the parameter attribute and assign the parameter vale
                        if param_declaration:
                            # Add param statement
                            piece_to_replace = dec_line.split('::')[0]
                            replacement = "%s, parameter " % piece_to_replace
                            dec_line = dec_line.replace(piece_to_replace, replacement)

                            # Add param assignation
                            dec_line = dec_line + "=" + parameter_value

                            # Substitute the line
                            self.lines[line_index - dec_line_index - 1] = dec_line

                            break

                # Delete parameter declaration
                self.lines[line_index] = ""

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:]])

    def save_modified_file(self, filepath):
        with open(filepath, "w") as output_file:
            output_file.write(self.text)

    def pre_process(self, replace_reals=False, replace_wrk_alloc=False):
        # Remove all comments from all the lines
        self.remove_all_comments()
        # Deletes all the empty lines
        self.lines = remove_empty_lines(self.lines)
        # Merges commands split on multiples lines
        self.lines =merge_lines(self.lines)
        # Split multiple command lines
        self.split_lines()
        # Split lines with more than one declaration
        self.split_declarations()

        # Just needed for old Nemo version
        if replace_wrk_alloc:
            self.replace_workalloc_in_file()

        # remove parameter assignation F77 style
        self.remove_F77_parameter_declaration()
        # Put text back together
        self.rebuild_text()
        if replace_reals:
            # Replace real declarations with type(rpe_var)
            self.replace_real_declaration()
            # Workaround for parameters
            self.replace_parameter_declaration()

    def post_process(self):
        # If we don't have source info there's nothing to process
        if self.source_info is None:
            raise AssertionError("It is needed to have a vault database to postprocess the code")

        # Process intrinsic calls
        IntrinsicDealer.fix_intrinsic_calls(self)
        # Prosess implicict function
        self.fix_implicit_functions()
        # Process routine calls
        self.fix_routines()
        # Process function calls
        self.fix_functions()

        DerivedTypeDealer.fix_type_conversion(self)

        self.fix_data_statement()

        # Process read and write calls
        self.fix_read_and_write_calls()
        # Fix pointer assignations
        PointerDealer.fix_pointer_assignations(self)
        # Add use rp_emulator statement
        self.add_use_rpe_statement()


    # Look at each line for REAL variable declarations and replace it with type(rpe_var)
    def replace_real_declaration(self):
        for _index, line in enumerate(self.lines):
            _match = re.search('[^\!]*(real).*(::|(function))', line, flags=re.I)
            if _match:
                _match_real_with_prec = re.search('[^\!]*(real *\([\w= ]*\)).*(::|(function))', line, flags=re.I)
                if _match_real_with_prec:
                    real_type = _match_real_with_prec.groups(0)[0]
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
                else:
                    _match = re.search('^\s*\w?(real).*(::|(function))', line, flags=re.I)
                    if _match:
                        real_type = _match.groups(0)[0]
                    else:
                        raise Exception
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
        self.rebuild_text()

    def replace_parameter_declaration(self):
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(self.lines):
            if re.search('^[^\!]*type\(rpe_var\).*::.*=', line, flags=re.I):
                arguments = line.split("::")[1]

                type_of_exception = self.find_kind_of_parameter_exception(arguments)

                # The assignation is made through a real number
                if type_of_exception == 'is_a_number':
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                # The assignation is made with a RPE variable
                elif type_of_exception == "regular":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # take the real part of the RPE var
                    piece_to_replace_real = " %s%%val " % piece_to_replace
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace_real
                    self.lines[_index] = replace_variable_exact_match(piece_to_replace, replacement, line)
                elif type_of_exception == "multiple_variables":
                    for argument in arguments.split(","):
                        piece_to_replace = argument.split("=")[1].strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == "array_initialization":
                    # Keep just the right part of  assignation avoid error when dimensions are specified
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = add_rpe_var_to_arguments(piece_to_replace)

                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                elif type_of_exception == "array_reshape":
                    array = arguments.split("[")[1]
                    array = array.split("]")[0]

                    # List the unique arguments inside the reshape function
                    arguments_in_reshape = list(set([a.strip() for a in array.split(",")]))

                    # Add the rpe_var cast into each one of them
                    for arg in arguments_in_reshape:
                        piece_to_replace = arg.strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == "has_operation":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # Get the elements in operation
                    split_elements = split_elements_in_operations(piece_to_replace)
                    replacement = " rpe_var(  52, %s ) " % piece_to_replace
                    for element in split_elements:
                        operand = element.strip()

                        # If is a real
                        if not is_a_number(operand):
                            operands_real_part = " %s%%val " % operand
                            replacement = replace_variable_exact_match(operand, operands_real_part, replacement)
                    line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == 'has_intrinsics':

                    # Get the name of the intrinsic to split the line
                    intrinsic_name = IntrinsicDealer.has_intrinsics(arguments, True)
                    piece_to_replace = arguments.lower().split(intrinsic_name)[1].strip()

                    # Match the name of the RPE variable inside the parenthesis
                    p = re.search('\((.*?)\)', piece_to_replace)
                    piece_to_replace = p.group(1)

                    # Get the real part of the RPE variable
                    replacement = " %s%%val " % piece_to_replace

                    # Substitute the mach back in the line
                    line = line.replace(piece_to_replace, replacement)
                    piece_to_replace = line.split("=")[1].strip()

                    # Add the cast to RPE var
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace

                    # Save changes
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                    # self.lines[_index] = line
                elif type_of_exception == "already_fixed":
                    continue
        self.rebuild_text()

    def fix_data_statement(self):
        changed_something = False
        current_block = CurrentBlock(self.module.main())
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(self.lines):
            match = re.search(r'^( *)data +(.*) +/(.*)/', line, flags=re.I)
            if match:
                spaces, var_name, values = match.groups(0)
                if var_name.count(","):
                    var_type = get_type_of_contents(var_name.split(",")[0], routine=current_block.routine,
                                                vault=self.source_info)
                else:
                    var_type = get_type_of_contents(var_name, routine=current_block.routine, vault=self.source_info)
                if var_type != "rpe_var":
                    continue
                values = ",".join(["rpe_var(52, %s)" % x for x in values.split(",")])
                self.lines[_index] = "%sdata %s /%s/" % (spaces, var_name, values)
                changed_something = True
            else:
                current_block = current_block.in_which_block(line, vault=self.source_info)
        if changed_something:
            self.rebuild_text()

    def add_use_rpe_statement(self):
        # Adding module rp_emulator (in theory adding this to par_kind.f90 should be enough,
        # but theory fails and compiler complains so I'll add it everywhere
        self.lines.insert(1, "USE rp_emulator")
        self.rebuild_text()

    def fix_read_and_write_calls(self):
        self.fix_read_calls()
        self.fix_write_calls()

    def fix_read_calls(self):
        current_block = CurrentBlock(self.module.main())
        pattern = "[^\w](read *\(.*)"
        for line_index, line in enumerate(self.lines):
            matches = re.search(pattern, line, re.IGNORECASE)
            if re.search("write *\(", line, re.I):
                continue
            if matches:
                read_call = close_brackets(matches.groups(1)[0])
                # Remove the read call
                read_arguments = line[matches.start(1):].replace(read_call, "")
                # Continue just if arguments follow the call
                if not read_arguments:
                    continue
                replacement = read_arguments.replace(" ", "")
                arguments = split_read_arguments(replacement)

                # Check the argument type and replace just if is an rpe
                for argument_index, argument in enumerate(arguments):
                    argument_type = get_type_of_contents(argument, routine=current_block.routine, vault=self.source_info)
                    if argument_type == "rpe_var":
                        replacement = replace_variable_exact_match(argument, argument + "%val", replacement)

                # Rebuilt the line
                # self.lines[line_index] = read_call + " " + read_arguments
                self.lines[line_index] = line.replace(read_arguments, replacement)
            else:
                current_block = current_block.in_which_block(line, vault=self.source_info)

    def fix_write_calls(self):
        current_block = CurrentBlock(self.module.main())
        pattern = "^[^\!]*\s*write\([^()]*\) (.*)."

        for line_index, line in enumerate(self.lines):
            if re.search("WRITE *\(", line):
                wc = close_brackets(line[line.find("WRITE"):])

                line = line.replace(wc, "WRITE()")
            else:
                current_block = current_block.in_which_block(line, vault=self.source_info)
            matches = re.search(pattern, line.strip(), re.IGNORECASE)
            if matches:
                line_container, arguments = split_line_container_and_arguments(self.lines[line_index])
                if not arguments:
                    continue
                mask = literal_mask(line)
                for argument_index, argument in enumerate(arguments):
                    if mask[line.index(argument) + 1]:
                        continue
                    argument_type = get_type_of_contents(argument, routine=current_block.routine, vault=self.source_info)
                    if argument_type == "rpe_var":
                        arguments[argument_index] = "real( %s )" % argument
                replacement = ",".join(arguments)
                self.lines[line_index] = line_container.replace("###TOREPLACE###", replacement)

    def fix_routines(self):
        current_block = CurrentBlock(self.module.main())
        exceptions = {}
        temporal_variable_counter = Counter()
        for _index, _line in enumerate(self.lines):
            # Find routine calls in line
            routine_calls = calls_to_routines(_line)
            if not routine_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = current_block.in_which_block(_line, vault=self.source_info)
            else:
                temporal_line = _line
                for call in routine_calls:
                    is_external = False
                    called_routine = find_called_routine(call)
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    # In case the procedure is not found, consider it external
                    try:
                        procedure = self.source_info.get_procedure_by_name(called_routine)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = get_interface_from_arguments(called_arguments, interface=procedure,
                                                                     routine=current_block.routine,
                                                                     vault=self.source_info)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_routine)
                        is_external = True

                        # continue

                    temporal_call = call
                    if not is_external:
                        for argument_index, argument in enumerate(called_arguments):
                            if is_keyword_argument(argument):
                                [keyword, val] = argument.lower().split("=", 1)
                                argument_type = get_type_of_contents(val.strip(), routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword.strip())
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in VariablePrecision.real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in VariablePrecision.real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = get_type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in VariablePrecision.real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        try:
                                            exceptions[_index].append(
                                                (argument, argument_index, dummy_argument, current_block.routine))
                                        except KeyError:
                                            exceptions[_index] = [
                                                (argument, argument_index, dummy_argument, current_block.routine)]
                                elif argument_type == "rpe_var" and dummy_argument.type in VariablePrecision.real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "inout":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = get_type_of_contents(argument, routine=current_block.routine,
                                                             vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                if re.match("^\w+$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif re.match("^\w+ *\([^\(\)]*\)$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif not has_operations(argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                else:
                                    temporal_call = add_real_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[_index] = temporal_line
        self.rebuild_text()

        # Manage exceptions
        added_lines = 0
        for line in sorted(exceptions):
            for argument, argument_index, dummy_argument, routine in exceptions[line]:
                current_line = line + added_lines
                use_temporal_variable(self, current_line, argument, argument_index, routine=routine,
                                      dummy=dummy_argument, counter=temporal_variable_counter)
                added_lines += 3
            added_lines += 2 * len(exceptions[line])
        self.rebuild_text()

    def fix_functions(self):
        current_block = CurrentBlock(self.module.main())
        for line_index, line in enumerate(self.lines):
            function_calls = calls_to_functions(line, routine=current_block.routine, vault=self.source_info)
            if not function_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = current_block.in_which_block(line, vault=self.source_info)
            else:
                # Fix function calls (line_index
                temporal_line = line
                for call in function_calls:
                    # Small workaround to avoid identifying routines as functions.
                    patt = "call +%s" % call
                    patt = patt.replace("(", r"\(")
                    patt = patt.replace(")", r"\)")
                    patt = patt.replace("/", r"\/")
                    patt = patt.replace("*", r"\*")
                    m = re.search(patt, line, re.I)
                    if m:
                        continue
                    is_external = False
                    called_function = find_called_function(call)

                    if called_function.lower() == current_block.routine.name.lower():
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    try:
                        procedure = self.source_info.get_procedure_by_name(called_function)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = get_interface_from_arguments(called_arguments, interface=procedure,
                                                                     routine=current_block.routine,
                                                                     vault=self.source_info)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_function)
                        is_external = True

                    if not is_external:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides. sbcice_lim and lib_fortran fail without it
                            if argument == ":":
                                continue
                            if is_keyword_argument(argument):
                                keyword = argument.split("=")[0].strip()
                                argument_type = get_type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in VariablePrecision.real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in VariablePrecision.real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = get_type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in VariablePrecision.real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in VariablePrecision.real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "inout":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = get_type_of_contents(argument, routine=current_block.routine,
                                                             vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                temporal_call = add_val_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[line_index] = temporal_line
        self.rebuild_text()

    def fix_implicit_functions(self):
        try:
            current_block = CurrentBlock(self.module.main())
        except AttributeError:
            return None
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        pattern = r"^ *(\w+) *\([\w, ]*\) *= *(.*)"
        # Look into each line
        for _index, line in enumerate(self.lines):
            m = re.search(pattern, line)
            if m:
                routine = current_block.routine
                call = close_brackets(line[m.start(1):])
                call_arguments = find_call_arguments(call)
                is_function = False
                for argument in call_arguments:
                    argument_type = get_type_of_contents(argument, routine=routine, vault=self.source_info)
                    if argument_type != "integer":
                        is_function = True
                        break

                if is_function:
                    function_name = find_called_function(call)
                    new_function = Function(function_name, module=self.module)
                    new_function.type = get_type_of_contents(function_name, routine=routine, vault=self.source_info)
                    for argument_index, argument in enumerate(call_arguments):
                        var = routine.get_variable_by_name(argument)
                        var.is_dummy_argument = True
                        var.position = argument_index
                        var.intent = "in"
                        new_function.add_variable(var)

                    self.insert_function(new_function, _index)
                    # Now remove var from vault mutable variables
                    for var in self.source_info.variables:
                        name = var.name
                        module_name = var.routine.module.name
                        routine_name = var.routine.name
                        if name == function_name and module_name == new_function.module.name and routine_name == routine.name:
                            self.source_info.variables.remove(var)
                            break

            else:
                current_block = current_block.in_which_block(line, vault=self.source_info)

    # Transform an implicit function in a normal function
    def insert_function(self, new_function, line_index):

        function_declaration, function_definition = self.lines[line_index].split("=")

        # Now that we stored the function we can delete its usage
        self.lines[line_index] = ""

        # Now we proceed to remove the variables declaration and the function declaration
        n_var = len(new_function.variables) + 1
        for _index, _line in enumerate(reversed((self.lines[:line_index]))):
            if n_var == 0:
                break
            if _line.count("::"):
                declared_var = _line.split("::")[1].strip()
                for var in new_function.variables:
                    if var.name == declared_var or new_function.name == declared_var:
                        self.lines[line_index - _index - 1] = ""
                        n_var -= 1
                        break
        # Now remove these lines
        self.lines = remove_empty_lines(self.lines)

        # Search for a line to insert the new function definition
        fun_end = ObtainSourcefileInfo.find_next_end(self, line_index, "end subroutine")
        sbr_ned = ObtainSourcefileInfo.find_next_end(self, line_index, "end function")
        insertion_line = min(fun_end, sbr_ned) + 1
        # Insert the function

        self.lines.insert(insertion_line, "END FUNCTION")
        function_definition = "x =" + function_definition
        self.lines.insert(insertion_line, function_definition)
        if new_function.type == "rpe_var":
            ret_val_declaration = "type(" + new_function.type + ") :: x"
        else:
            ret_val_declaration = new_function.type + " :: x"
        self.lines.insert(insertion_line, ret_val_declaration)
        for var in new_function.variables:
            if var.type == "rpe_var":
                var_declaration = "type(" + var.type + ") :: " + var.name
            else:
                var_declaration = var.type + " :: " + var.name
            self.lines.insert(insertion_line, var_declaration)

        function_declaration = "FUNCTION" + function_declaration + "result(x)"
        self.lines.insert(insertion_line, function_declaration)

        self.rebuild_text()

    def find_kind_of_parameter_exception(arguments):
        if not arguments.count(","):
            argument = arguments.split("=")[1]
            num_type = is_a_number(arguments)
            if num_type:
                return "is_a_number"
            if has_operations(argument):
                type_of_exception = 'has_operation'
            elif has_intrinsics(argument):
                type_of_exception = 'has_intrinsics'
            else:
                type_of_exception = "regular"
        else:
            if arguments.count("=") > 1:
                type_of_exception = "multiple_variables"
            elif arguments.count("(/"):
                type_of_exception = "array_initialization"
            elif arguments.count("RESHAPE(["):
                type_of_exception = "array_reshape"
            elif re.search(r"= *\[.*\]", arguments):
                type_of_exception = "array_reshape"
            elif arguments.count("rpe_var("):
                type_of_exception = "already_fixed"
            else:
                raise TypeError("Unexpected kind of exception")
        return type_of_exception

