from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.BasicFunctions import close_brackets
from AutoRPE.UtilsRPE.BasicStructures import Routine, Interface
from AutoRPE.UtilsRPE.SourceManager import match_module_and_filename, expand_sources,\
    load_source_files, load_vault, load_variable_list
from AutoRPE.UtilsRPE.Cleaner import remove_comments
from AutoRPE.UtilsRPE.AttributeParser import get_variable_name
from AutoRPE.UtilsRPE.CallManager import find_argument_intervals, find_called_routine, find_call_arguments, \
    is_call_to_function, calls_to_routines
from AutoRPE.UtilsRPE.NatureDeterminer import has_operations
from AutoRPE.UtilsRPE.Finder import find_file_and_line_of_allocation, \
    find_file_and_line_of_namelist_reading, find_first_use, find_declaration_line
from AutoRPE.UtilsRPE.Splitter import split_vars_by_allocatable, split_by_appearance_in_namelist, \
    split_vars_by_main_or_routine
from AutoRPE.UtilsRPE.Getter import get_variable, get_interface_from_arguments, is_keyword_argument, \
    get_type_of_contents
from AutoRPE.UtilsRPE.Counter import Counter
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import warnings
import re
import pickle
from os.path import join, isfile


def insert_read_precisions_module_to_file(path_to_sources, target_module=None, vault=None):
    from os.path import dirname, abspath
    print("\n\nCreating module to read precision namelist...")
    # Adding the use statement in all the files.
    line_to_add = "   USE read_precisions\n"
    source_files = load_source_files(path_to_sources)

    for name, source_file in source_files.items():
        source_file.lines.insert(1, line_to_add)
        source_file.rebuild_text()
        source_file.save_modified_file(join(path_to_sources, source_file.filename))

    # Create module
    rpe_vars = [var for var in vault.variables if var.type == "rpe_var"]
    assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
    ids = [var.id for var in assigned_rpe_vars]
    # Path to template is hardcoded
    package_directory = dirname(abspath(__file__))
    filename = package_directory + "/../AdditionalFiles/read_precisions_template"

    with open(filename, "r") as input_file:
        template = input_file.read()
    module = template.replace("%NUMBER_OF_RPE_VARS%", "%i" % max(ids))
    with open(join(path_to_sources, "read_precisions.f90"), "w") as output_file:
        output_file.write(module)
    # Put call into target module

    if target_module:
        if isfile(join(path_to_sources, target_module)):
            with open(join(path_to_sources, target_module), "r") as input_file:
                original_source = input_file.read()

            lines = original_source.split("\n")
            lines.insert(1, "use read_precisions")
            for index, line in enumerate(lines):
                if line.lower().strip() == "private":
                    lines.pop(index)
            original_source = "\n".join(lines)

            with open(join(path_to_sources, target_module), "w") as output_file:
                output_file.write(original_source)
        else:
            warnings.warn("File %s not found in sources.\nUnable to insert read_precisions module")


def insert_variable_precision_after_routine_call(source_files, vault, counter):
    for name, source_file in source_files.items():
        try:
            current_block = CurrentBlock(vault[source_file.filename.replace(".f90", "")].main())
        except AttributeError:
            warnings.warn("No module %s" % source_file.filename)
            return

        for _index, _line in enumerate(source_file.lines):
            # Find routine calls in line
            routine_calls = calls_to_routines(_line)
            variables_to_truncate = []
            if not routine_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = current_block.in_which_block(_line, vault=vault)
            else:
                # Keep the condition if exists
                if _line.lower().strip().find("if(") == 0:
                    condition = close_brackets(_line)

                else:
                    condition = None
                # If there's some call look at each one of them
                for call in routine_calls:
                    is_external = False
                    called_routine = find_called_routine(call)
                    # If the routine is apply_truncation, continue
                    if called_routine == "apply_truncation":
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        # If there's no argument, continue
                        continue

                    # Check that we properly found the routine in the database
                    try:
                        # Find the procedure called (object)
                        procedure = vault.get_procedure_by_name(called_routine)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = get_interface_from_arguments(called_arguments, interface=procedure,
                                                                     routine=current_block.routine, vault=vault)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_routine)
                        is_external = True

                    # If the routine is in the database
                    if not is_external:
                        # Check each argument
                        for argument_index, argument in enumerate(called_arguments):
                            if argument.count("tmp_var_"):
                                # Skipping temporal variables
                                continue
                            # Look if argument is keyword argument
                            if is_keyword_argument(argument):
                                # In case it is a keyword argument, get the dummy argument using the name.
                                keyword = argument.split("=")[0].strip()
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Keyword variable not found: %s" % keyword)

                                if dummy_argument.type != "rpe_var":
                                    # If the argument is not an rpe_var because it's passing the %val element
                                    # of the struct it should be truncated anyway.
                                    if dummy_argument.type in VariablePrecision.real_types and argument.count("%val"):
                                        variables_to_truncate.append(argument.replace("%val", "").strip())
                                elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                    variables_to_truncate.append(argument)
                            else:
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if dummy_argument.type != "rpe_var":
                                    if dummy_argument.type == VariablePrecision.real_types and argument.count("%val"):
                                        # If the argument contains whichever operations, for sure is intent in
                                        is_c2f = is_call_to_function(argument)
                                        has_op = has_operations(argument)
                                        if not is_c2f and not has_op:
                                            argument = argument.split("%val", 1)[0]
                                            if argument:
                                                variables_to_truncate.append(argument.strip())
                                            else:
                                                raise AssertionError
                                elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                    variables_to_truncate.append(argument)
                    else:
                        # If the routine is external
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            argument_type = get_type_of_contents(argument, routine=current_block.routine, vault=vault)
                            if argument_type == "rpe_var":
                                variables_to_truncate.append(argument)
                            elif argument_type in VariablePrecision.real_types and argument.count("%val"):
                                variables_to_truncate.append(argument.replace("%val", "").strip())
            i = _index
            for variable in variables_to_truncate:
                i = insert_in_line(variable, source_file, current_block.routine, counter, i, vault=vault,
                                   condition=condition)

        source_file.rebuild_text()
        return


def insert_variable_precision_specification(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    # Add precision specification
    new_line = condition + "    %s%%sbits = emulator_variable_precisions(%i)" % (variable.name, variable.id)
    sourcefile.lines.insert(insertion_line, new_line)

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_usage(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_truncation(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    new_line = condition + "    CALL apply_truncation(%s)" % variable.name
    sourcefile.lines.insert(insertion_line, new_line)
    sourcefile.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, file_objects, counter_object):
    sourcefile, allocation_line = find_file_and_line_of_allocation(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, allocation_line + 1)
        insert_variable_usage(variable, sourcefile, allocation_line + 1)


def insert_variable_precision_specification_to_namelist_parameter(variable, file_objects, counter_object):
    sourcefile, reading_line = find_file_and_line_of_namelist_reading(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, reading_line + 1)
        insert_variable_usage(variable, sourcefile, reading_line + 1)
    else:
        warnings.warn("Variable %s not found" % variable.name)


def insert_variable_precision_to_routine_variable(variable, file_objects, counter_object):
    sourcefile, first_use_line = find_first_use(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        _line = sourcefile.lines[first_use_line]
        if re.search(r'^\s*if\s*\(', _line, re.I):
            condition = close_brackets(_line)
        else:
            condition = None
        insert_variable_precision_specification(variable, sourcefile, first_use_line, condition=condition)
        insert_variable_truncation(variable, sourcefile, first_use_line + 1, condition=condition)
        insert_variable_usage(variable, sourcefile, first_use_line + 2, condition=condition)
    else:
        warnings.warn("Variable %s not found %s in routine %s" % (
            variable.name, variable.routine.module.name, variable.routine.name))


def insert_in_line(var, source_file, routine, counter, line_index, vault=None, condition=None):
    inner_index = line_index
    var_object = get_variable(var, routine=routine, vault=vault)
    # Need to solve parameter issue here
    if var_object.is_parameter:
        return inner_index
    elif var_object.intent == "in":
        return inner_index

    if condition is None:
        condition = ""
    else:
        condition = condition

    sbits_assignation = "    %s%%sbits = emulator_variable_precisions(%i)"
    apply_truncation = "    CALL apply_truncation(%s) "

    if var_object.id is None:
        var_object.id = counter.count
        var_object.mutable = True
        counter.up()
    if var.count('='):
        var = var.split('=')[1].strip()

    source_file.lines.insert(inner_index + 1, condition + sbits_assignation % (var, var_object.id))
    source_file.lines.insert(inner_index + 2, condition + apply_truncation % var)

    return inner_index + 2


def add_sbits_to_rpe_variables(path_to_sources, vault=None):
    print("\n\nAssigning significant bits...")

    def condition(v):
        return v.type == "rpe_var" and not v.is_parameter and not v.name.count("tmp_var_")

    source_files = load_source_files(path_to_sources)
    rpe_variables = [var for var in vault.variables if condition(var) and not var.dimension == None]
    allocatable, non_allocatable = split_vars_by_allocatable(rpe_variables)
    var_counter = Counter()
    # Allocatables:
    for variable in allocatable:
        insert_variable_precision_specification_to_allocatable(variable, source_files, var_counter)

    # Non allocatables that appear in a namelist:
    na_in_namelist, na_not_in_namelist = split_by_appearance_in_namelist(non_allocatable, source_files)
    for variable in na_in_namelist:
        insert_variable_precision_specification_to_namelist_parameter(variable, source_files, var_counter)
    in_main, not_in_main = split_vars_by_main_or_routine(na_not_in_namelist)

    # Variables inside routines
    for variable in not_in_main:
        if variable.intent != "in":
            insert_variable_precision_to_routine_variable(variable, source_files, var_counter)

    # Adding sbits assignation and apply_truncation after routine calls
    insert_variable_precision_after_routine_call(source_files, vault=vault, counter=var_counter)

    # To save the variable id into the vault.
    save = True
    if save:
        # Saving modified files
        for filename, source_file in source_files.items():
            source_file.save_modified_file(join(path_to_sources, filename))
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return


def add_function_to_argument(string, field_index, _function, number_of_fields=None):
    # Given a line with a call, add a call to a function in a specific field by index.
    num_arguments, interval = find_argument_intervals(string, field_index)
    if num_arguments <= field_index:
        return string
    if number_of_fields:
        if number_of_fields is not num_arguments:
            return string
    first_segment = string[:interval[0]]
    intermediate_segment = string[interval[0]:interval[1]]
    last_segment = string[interval[1]:]

    # The arguments containing a literal are discarded (Probably a bad workaround)
    if intermediate_segment.count("'"):
        #     return _line
        # raise ValueError("Probably the argument should not contain quotes at this point")
        warnings.warn("Probably the argument should not contain quotes at this point:\n%s" % intermediate_segment)

    # The arguments that already contain a call to the function are discarded (Probably a bad workaround)
    if re.match(r"^ *%s *\(.*" % _function, intermediate_segment):
        # raise ValueError("Probably the argument should not contain %s at this point" % _function)
        return string

    # If there are named arguments, add the function to the actual argument
    # (if the equal sign is inside a parenthesis, this should not be done)
    named_argument_workaround = False

    if intermediate_segment.count("="):
        if intermediate_segment.count("("):
            if intermediate_segment.find("=") < intermediate_segment.find("("):
                named_argument_workaround = True
        else:
            named_argument_workaround = True

    if named_argument_workaround:
        parts = intermediate_segment.split("=")
        intermediate_segment = "%s=%s( %s )" % (parts[0], _function, parts[1])
    else:
        intermediate_segment = "%s( %s )" % (_function, intermediate_segment)
    string = "%s%s%s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_val_to_argument(string, _field):
    num_arguments, intervals = find_argument_intervals(string)
    for i in [_field]:
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        if intermediate_segment.strip()[:2] == "(/":
            intermediate_segment = intermediate_segment.strip()[2:]
            first_segment = first_segment + "(/"
        if intermediate_segment.strip()[-2:] == "/)":
            intermediate_segment = intermediate_segment.strip()[:-2]
            last_segment = "/)" + last_segment

        string = "%s %s%%val %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_rpe_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "rpe", number_of_fields)
    return string


def add_rpe_var_to_arguments(string):
    num_arguments, intervals = find_argument_intervals(string)
    for i in range(num_arguments):
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        string = "%s rpe_var( 52, %s ) %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_real_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "real", number_of_fields)
    return string


def replace_variable_exact_match(piece_to_replace, replacement, str):
    # In case our string does not start with a - or +, we'll add the \b to indicate a word border
    if piece_to_replace.strip()[0] not in ["-", "+"]:
        piece_to_replace = "\\b" + piece_to_replace.strip()
    # Adapt the string to fulfill regex rules
    piece_to_replace = piece_to_replace.replace("+", r"\+")
    piece_to_replace = piece_to_replace.replace("-", r"\-")
    piece_to_replace = piece_to_replace.replace("*", r"\*")
    if piece_to_replace.count("("):
        piece_to_replace = piece_to_replace.replace("(", "\s*\(")
        # Avoid matching structures
        if not piece_to_replace.count("%"):
            piece_to_replace = piece_to_replace.replace(")", "\)(?!%)")
        else:
            piece_to_replace = piece_to_replace.replace(")", "\)")
    else:
        piece_to_replace = piece_to_replace + "\\b(?!%val)"

    # Replace string using regex
    str = re.sub(piece_to_replace, replacement, str)
    return str


# This function will find the file where a variable is declared and will replace the declaration
# with a new one using the new precision spec.
def replace_declaration(variable, file_paths, vault, target_precision="dp"):
    files = file_paths
    module = variable.routine.module.name
    module_file_paths = match_module_and_filename(module, files)
    replacement_done = False
    for filepath in module_file_paths:
        try:
            additional_files = []
            declaration_line_indexes = find_declaration_line(variable, filepath, vault, additional_files)
        except Error.LineNotFound:
            if len(additional_files):
                for h90_file in additional_files:
                    h90_file_path = [file for file in file_paths if file.count(h90_file)]
                    module_file_paths.extend(h90_file_path)
            else:
                pass
            continue
        replacement_done = True
        with open(filepath, "r") as f:
            try:
                data = [l for l in f]
            except UnicodeDecodeError:
                raise UnicodeDecodeError("The file %s contains non UTF-8 character.\n "
                                         "Please clean, and retry.", filepath)
            for declaration_line_index in declaration_line_indexes:
                declaration_line = data[declaration_line_index]
                new_line = replace_real_declaration(variable, declaration_line, target_precision=target_precision,
                                                    filename=filepath)
                data[declaration_line_index] = new_line
            updated_text = "".join(data)
        with open(filepath, "w") as output:
            output.write(updated_text)
    if not replacement_done:
        print("Declaration of variable %s with id %i not found in module %s" % (
            variable.name, variable.id, variable.routine.module.name))


def change_code(list_filename, source_path, vault_path, target_precision="dp"):
    # The purpose of this function is to modify the declarations of several variables in a source-code.
    # As inputs it takes the path to the file with the list of variables, the path to the sources that will be modified
    # and the path to the vault database containing information about the code.
    #
    # This function takes the following inputs:
    # - list_filename: Path to a file that contains a list of variables that will be modified in the code
    # - source_path: Path to a folder that contains the source files of the code corresponding to the vault.
    # - vault_path: Path to a vault database.

    vault = load_vault(vault_path)
    variable_ids = load_variable_list(list_filename)
    file_paths = expand_sources(source_path=source_path)

    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]
    for variable in variables:
        replace_declaration(variable, file_paths, vault, target_precision=target_precision)


## PRIVATE

# This function takes a declaration line (string) and updates it with the new type
# If the declaration contains more variables it keeps the other variables with the original attributes.
def replace_real_declaration(variable, declaration_line, target_precision="dp", filename=None):
    declaration_line_without_comments = remove_comments(declaration_line).strip()

    attributes, arguments = declaration_line_without_comments.split("::")
    variable_name = get_variable_name(arguments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(variable_name) == 1:
        if declaration_line_without_comments[-1] == "&":
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
            print("WARNING!! the declaration of %s in %s hasn't been properly fixed!" % (variable.name, filename))
        else:
            m = re.search(pattern, declaration_line_without_comments, re.I)
            if m:
                real_statement = m.group(0)
                precision_specification = m.group(1)
                new_real_statement = real_statement.replace(precision_specification, target_precision)
                new_declaration_line_without_comments = \
                    declaration_line_without_comments.replace(real_statement, new_real_statement)
                new_declaration_line = \
                    declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
            else:
                warnings.warn("Problems finding real statement:\n%s" % declaration_line)
                new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                       declaration_line
        return new_declaration_line
    else:
        if variable.name not in variable_name:
            raise AssertionError("Variable %s not found in list of arguments" % variable.name, attributes)
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line
