from AutoRPE.UtilsRPE.BasicStructures import Interface
from AutoRPE.UtilsRPE.BasicFunctions import create_type_combinations
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ReadSourceFile
from AutoRPE.UtilsRPE.Getter import get_dimension_of_contents, get_type_of_contents,\
    get_interface_from_arguments_and_info, is_keyword_argument, get_variable
from AutoRPE.UtilsRPE.CallManager import calls_to_routines, calls_to_functions, \
    find_called_routine, find_called_function, find_call_arguments
from AutoRPE.UtilsRPE.Cleaner import clean_contents
from AutoRPE.UtilsRPE.NatureDeterminer import is_a_number
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import warnings
import networkx as nx
from os import listdir
from os.path import isfile, join


def track_dependencies(path_to_model_sources, vault, graph):
    # Find files in sources
    all_sources = [f for f in listdir(path_to_model_sources) if
                   isfile(join(path_to_model_sources, f)) and f.count(".f90")]
    # Process Sources!
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        print("\rTracing dependencies %i/%i  %20s" % (index + 1, total, filename), end="")
        track_dependencies_in_file(join(path_to_model_sources, filename), vault, graph)


def create_nodes(variables, graph):
    for v in variables:
        try:
            graph[var_identifier(v)]
        except KeyError:
            dim = v.dimension if v.dimension is not None else 0
            graph.add_node(var_identifier(v),
                           dimension=dim,
                           routine=str(v.routine.name),
                           module=str(v.routine.module.name),
                           intent=str(v.intent),
                           type=str(v.type))


def create_edges(argument, dummy, graph):
    v1, v2 = argument, dummy
    dim = v1.dimension if v1.dimension is not None else 0
    graph.add_edge(var_identifier(v1), var_identifier(v2),
                   routine=str(v2.routine.name),
                   kind=str(v2.intent),
                   weight=1 + dim / 4.
                   )


def manage_loc_exception(called_arguments, vault, routine):
    # When we have a mpp_XXXloc exception instead of linking arguments with dummies we'll link the arguments together
    called_vars = []
    for arg in called_arguments:
        try:
            var = get_variable(arg, routine=routine, vault=vault)
            called_vars.append(var)
        except Error.VariableNotFound:
            continue
    argument_variables = []
    dummies = []
    for index1 in range(len(called_vars)):
        for index2 in range(index1 + 1, len(called_vars)):
            argument_variables.append(called_vars[index1])
            dummies.append(called_vars[index2])
    return argument_variables, dummies


# Scan the line to retrieve all real numbers used in routine calls or pointers assignation, and corresponding targets
def retrieve_real_in_line(line, vault, current_block, check=False):
    # Vectors to store return values
    all_var = []
    all_dummy_var = []

    # If check mode, returns procedures to fix and their arguments
    procedure_tf = []
    arguments_proc_tf = []
    argument_index_tf = []
    dummy_var_type_tf = []
    dummy_var_intent_tf = []

    # To store calls and called procedures names
    procedure_calls = []
    called_procedures = []

    # Search for calls to subroutines
    routine_calls = calls_to_routines(line)
    function_calls = calls_to_functions(line, routine=current_block.routine, vault=vault)

    if check:
        null_return_val = [None, None, None, None, None]
    else:
        null_return_val = [None, None]

    if routine_calls:
        for routine_call in routine_calls:
            called_routine = find_called_routine(routine_call)
            if called_routine == "date_and_time":
                continue
            procedure_calls.append(routine_call)
            called_procedures.append(called_routine)
    elif function_calls:
        for function_call in function_calls:
            procedure_calls.append(function_call)
            called_procedures.append(find_called_function(function_call))
    # Search for pointer assignment and returns if found
    elif PointerDealer.is_pointer_assignment(line):
        ptr, target = PointerDealer.get_pointer_target(line, current_block.routine, vault)

        if [ptr, target] == [None, None] or ptr.type not in ["single", "double", "rpe_var"]:
            return null_return_val
        if ptr.type != target.type:
            message = "POINTER => " + ptr.name + " is " + ptr.type + " instead of  " + target.type + \
                      "in procedure" + current_block.routine.name + "in file: " + current_block.module_name + "f90"
            if check:
                print(message)
            else:
                raise Error.PointerTarged(message)
        all_var.append(ptr)
        all_dummy_var.append(target)
    # Nothing found
    else:
        return null_return_val

    # If are there calls to procedures
    if called_procedures:
        for procedure_call, called_procedure in zip(procedure_calls, called_procedures):
            called_arguments = find_call_arguments(procedure_call)
            if called_procedure in ["mpp_minloc", "mpp_maxloc"]:
                v, av = manage_loc_exception(called_arguments, vault, routine=current_block.routine)
                all_var += v
                all_dummy_var += av
                continue
            if not called_arguments:
                continue
            try:
                procedure = vault.get_procedure_by_name(called_procedure, module_name=current_block.module_name)
            except Error.ProcedureNotFound:
                continue
            indices_to_ignore = []

            if isinstance(procedure, Interface):
                argument_types = [get_type_of_contents(argument, routine=current_block.routine, vault=vault) for argument in
                                  called_arguments]
                argument_dimensions = [get_dimension_of_contents(argument, routine=current_block.routine, vault=vault) for argument
                                       in called_arguments]
                keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in
                                     called_arguments]
                # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                # because there's already an existing version within the same interface with a different precision
                alternative_procedures = []
                for _in, _type in enumerate(argument_types):
                    new_argument_types = argument_types[:]
                    if _type == "single":
                        new_argument_types[_in] = "double"
                    elif _type == "double":
                        new_argument_types[_in] = "single"
                    else:
                        continue
                    try:
                        pp = get_interface_from_arguments_and_info(new_argument_types, procedure,
                                                                   argument_dimensions=argument_dimensions,
                                                                   keyword_arguments=keyword_arguments, vault=vault)
                        alternative_procedures.append(pp)
                        if not check:
                            indices_to_ignore.append(_in)
                    except AssertionError:
                        continue
                # Classical match
                try:
                    procedure = get_interface_from_arguments_and_info(argument_types, procedure,
                                                                      argument_dimensions=argument_dimensions,
                                                                      keyword_arguments=keyword_arguments, vault=vault)
                except AssertionError:
                    if check and not alternative_procedures:
                        # In case there wasn't a match with the interface, check if through a variable cast it can work
                        types_to_test = create_type_combinations(argument_types)
                        matches = []
                        for t in types_to_test:
                            try:
                                m = get_interface_from_arguments_and_info(
                                                                              t,
                                                                              procedure,
                                                                              argument_dimensions=argument_dimensions,
                                                                              keyword_arguments=keyword_arguments,
                                                                              vault=vault,
                                                                              )
                                matches.append(m)
                                # break
                            except AssertionError:
                                continue
                        if len(matches) == 1:
                            procedure = matches[0]
                        elif len(matches) > 1:
                            print("Multi-match!")
                        else:
                            print("No match!")
                        if isinstance(procedure, Interface):
                            print("INTERFACE: ", current_block.routine.name, procedure.name,
                                  [at for at in argument_types], "file: ")
                            continue
                    elif check and alternative_procedures:
                        procedure = alternative_procedures[0]
                    else:
                        # TODO: This workaround here was required because the function dimension_of_contents still
                        # does not work properly.
                        # TODO: Probably not needed anymore
                        try:
                            procedure = get_interface_from_arguments_and_info(argument_types, procedure,
                                                                              argument_dimensions=None,
                                                                              keyword_arguments=keyword_arguments, vault=vault)
                        except AssertionError:
                            raise AssertionError
            if procedure is None:
                continue
            for index, argument in enumerate(called_arguments):
                if not check:
                    if index in indices_to_ignore:
                        continue
                key = None
                if is_keyword_argument(argument):
                    [key, argument] = argument.split("=", 1)
                    key = key.strip()
                    argument = argument.strip()
                    var_type = get_type_of_contents(argument, routine=current_block.routine, vault=vault)
                else:
                    var_type = get_type_of_contents(argument, routine=current_block.routine, vault=vault)

                if var_type not in ["single", "double", "rpe_var"] or is_a_number(argument):
                    continue

                if key is not None:
                    dummy_var = procedure.get_variable_by_name(key)
                else:
                    dummy_var = procedure.get_variable_by_position(index)
                # In case of the argument being an input, avoid binding the variables.
                if not check and dummy_var.intent == "in":
                    continue
                if not check:
                    try:
                        var = get_variable(argument, routine=current_block.routine, vault=vault)
                        all_var.append(var)
                        all_dummy_var.append(dummy_var)
                    except AssertionError:
                        continue
                else:
                    if var_type != dummy_var.type:
                        var_name = clean_contents(argument)
                        print("ARGUMENT => ", var_name, " is ", var_type, " instead of  ", dummy_var.type,
                              "in call to", called_procedure, "file: ", current_block.routine.module.name, ".f90")
                        procedure_tf.append(called_procedure)
                        arguments_proc_tf.append(argument)
                        argument_index_tf.append(index)
                        dummy_var_type_tf.append(dummy_var.type)
                        dummy_var_intent_tf.append(dummy_var.intent)

    if check:
        return [procedure_tf, arguments_proc_tf, argument_index_tf, dummy_var_type_tf, dummy_var_intent_tf]
    else:
        return [all_var, all_dummy_var]


def track_dependencies_in_file(filename, vault, graph):
    try:
        current_block = CurrentBlock(vault[filename.split("/")[-1].replace(".f90", "")].main())
    except AttributeError:
        warnings.warn("No module %s" % filename)
        return
    f = ReadSourceFile(filename)

    for line in f.lines:
        # Returns all REAL var on the line and corresponding dummy argument
        var, dummy_var = retrieve_real_in_line(line, vault, current_block=current_block)
        if [var, dummy_var] != [None, None]:
            for v, dv in zip(var, dummy_var):
                # Create nodes
                create_nodes([v, dv], graph)
                # Add edges
                create_edges(v, dv, graph)
        else:
            current_block = current_block.in_which_block(line, vault=vault)


def propagate_dependencies(vault, graph=None):
    variables_in_graph = [v for v in vault.variables if var_identifier(v) in graph.nodes]
    var_dict = {}
    for v in variables_in_graph:
        var_dict[var_identifier(v)] = v
    undirected_graph = graph.to_undirected()
    for v in variables_in_graph:
        cc = nx.node_connected_component(undirected_graph, var_identifier(v))
        try:
            v.same_as = [var_dict[n] for n in cc]
        except KeyError:
            raise KeyError


def var_identifier(var):
    return "%s#%s#%s#%s" % (var.name, var.routine.name, var.module.name, str(var.id))
