from AutoRPE.UtilsRPE.LoadAndStore import load_intrinsics
from AutoRPE.UtilsRPE.BasicStructures import Routine, Function, Interface, ModuleMain
from AutoRPE.UtilsRPE.DerivedTypeDealer import DerivedType
import AutoRPE.UtilsRPE.Error as Error


class Vault:
    def __init__(self):
        self.modules = []
        self.variables = []
        self.dictionary_of_variables = {}
        self.procedures = []
        self.procedures_dictionary = None
        self.routines = []
        self.functions = []
        self.interfaces = []
        self.derived_types = []
        self.intrinsics = load_intrinsics()
        self.indexed_variables = None

    def add_module(self, module):
        self.modules.append(module)

    def __getitem__(self, item):
        for module in self.modules:
            if module.name == item:
                return module

    def process_procedures(self):
        # Purpose: Add the procedures defined in the modules to the list of procedures and classify by type.
        for module in self.modules:
            for procedure in module.procedures:
                if procedure not in self.procedures:
                    self.procedures.append(procedure)
        for procedure in self.procedures:
            if isinstance(procedure, Routine):
                if procedure not in self.routines:
                    self.routines.append(procedure)
            elif isinstance(procedure, Function):
                if procedure not in self.functions:
                    self.functions.append(procedure)
            elif isinstance(procedure, Interface):
                if procedure not in self.interfaces:
                    self.interfaces.append(procedure)
            elif isinstance(procedure, DerivedType):
                if procedure not in self.derived_types:
                    self.derived_types.append(procedure)

        self.procedures_dictionary = {"all": {}, "functions": {}, "subroutines": {}, "interfaces": {}, "types": {}}
        for procedure in self.procedures:
            if not isinstance(procedure, ModuleMain):
                # Determine which kind of procedure are we dealing with
                if isinstance(procedure, Function):
                    procedure_group = "functions"
                elif isinstance(procedure, Routine):
                    procedure_group = "subroutines"
                elif isinstance(procedure, Interface):
                    procedure_group = "interfaces"
                elif isinstance(procedure, DerivedType):
                    procedure_group = "types"
                else:
                    raise AssertionError("Procedure type not found")

                # Add procedure to dictionary, in its own section and in the dictionary with all elements
                try:
                    self.procedures_dictionary[procedure_group][procedure.name.lower()]
                    self.procedures_dictionary["all"][procedure.name.lower()]
                    procedure_arr = self.procedures_dictionary[procedure_group][procedure.name.lower()]
                    if procedure not in procedure_arr:
                        procedure_arr.append(procedure)
                        self.procedures_dictionary[procedure_group][procedure.name.lower()] = procedure_arr
                        self.procedures_dictionary["all"][procedure.name.lower()] = procedure_arr
                    else:
                        print("Procedure %s defined more than once" % procedure.name)

                except KeyError:
                    self.procedures_dictionary[procedure_group][procedure.name.lower()] = [procedure]
                    self.procedures_dictionary["all"][procedure.name.lower()] = [procedure]

    def process_variables(self):
        # Purpose: Add the variables defined in the modules to the list of variables.
        if not self.procedures:
            self.process_procedures()
        for procedure in self.procedures:
            if isinstance(procedure, ModuleMain) \
                    or isinstance(procedure, Function) \
                    or isinstance(procedure, Routine) \
                    or isinstance(procedure, DerivedType):
                for variable in procedure.variables:
                    self.variables.append(variable)
                    if not variable.name.lower() in self.dictionary_of_variables:
                        self.dictionary_of_variables[variable.name.lower()] = [variable]
                    else:
                        self.dictionary_of_variables[variable.name.lower()] += [variable]

    def index_variables(self):
        if not self.variables:
            self.process_variables()
        self.indexed_variables = {}
        for variable in self.variables:
            self.indexed_variables[variable.id] = variable

    def get_variable(self, var_name, routine=None):
        # Purpose: Retrieve a variable by name
        if not self.variables:
            self.process_variables()
        try:
            variables = []
            for variable in self.dictionary_of_variables[var_name.lower()]:
                variables.append(variable)
        except KeyError:
            pass
        # Just one var exists with this name
        if len(variables) == 1:
            return variables[0]
        elif len(variables) > 1 and routine is not None:
            var = [c for c in variables if c in routine.module.accessible_var +
                   routine.module.variables +
                   routine.variables]
            if len(var) == 1:
                return var[0]
            if not len(var):
                # Import are not done in a proper way, but use some sort of inheritance, I am not happy
                var = [c for c in variables if c in routine.module.indirect_accessible_var]
                if len(var) == 1:
                    return var[0]
                if not len(var):
                    raise Error.VariableNotFound
            # A variable has been declared both as global (routine.name == main ) and as private routine variable
            if len(var) > 1:
                shadow_var = [v for v in var if v.routine.name == routine.name]
                try:
                    return shadow_var[0]
                except IndexError:
                    print("here")
                print("A global variable has been shadowed " + var[0].name)
        else:
            raise Error.VariableNotFound

    def get_variable_by_id(self, var_id):
        # Purpose: Retrieve a variable by id

        # Few workarounds in case the vault does not have the indexed_variables
        try:
            self.indexed_variables
        except AttributeError:
            self.indexed_variables = None
        if not self.indexed_variables:
            self.index_variables()

        return self.indexed_variables[var_id]

    def get_function(self, item):
        # Purpose: Retrieve a function by name
        return self.get_procedure_by_name(item, procedure_type=Function)

    def get_procedure_by_name(self, item, procedure_type=None, module_name=None):
        # Purpose: Retrieve a procedure by name
        item = item.lower()
        # TODO: This routine is quite expensive (16%), could be boosted using dictionaries.
        if not self.procedures or not self.procedures_dictionary:
            self.process_procedures()

        new_approach = True

        if new_approach:
            if procedure_type is None:
                procedure_group = "all"
            elif procedure_type == Function:
                procedure_group = "functions"
            elif procedure_type == Routine:
                procedure_group = "subroutines"
            elif procedure_type == Interface:
                procedure_group = "interfaces"
            elif procedure_type == DerivedType:
                procedure_group = "types"
            else:
                raise AssertionError("Procedure type not found")

            try:
                procedure = self.procedures_dictionary[procedure_group][item]
                if len(procedure) > 1 and module_name is not None:
                    # this is safe until they start to define functions inside functions
                    return [proc for proc in procedure if proc.module.name == module_name][0]

                return self.procedures_dictionary[procedure_group][item][0]
            except KeyError:
                raise Error.ProcedureNotFound("Procedure with name %s not found" % item)

        else:
            for procedure in self.procedures:
                if procedure_type:
                    if procedure.name == item:
                        if isinstance(procedure, procedure_type):
                            return procedure
                elif procedure.name == item:
                    return procedure
            raise Error.ProcedureNotFound("Procedure with name %s not found" % item)

    def get_derived_type(self, item):
        # Purpose: Retrieve a derived type by name
        for dt in self.derived_types:
            if dt.name.lower() == item.lower():
                return dt

    def expand_use_statement(self):
        for module in self.modules:
            # use statement already stored
            used_modules = [m.uses for m in self.modules if m.name == module.name][0]

            # Store use statement of used modules
            additional_used_modules = []
            for um in used_modules:
                for var in um.variables:
                    module.accessible_var.append(var)
                if um.uses:
                    additional_used_modules.extend(um.uses)

            # Remove duplicates
            additional_used_modules = set(additional_used_modules)

            # Remove input already in used statement
            additional_used_modules = [aum for aum in additional_used_modules if aum not in module.uses]

            # Look recursively at used modules
            while not sorted([aum.name for aum in additional_used_modules]) == sorted([mu.name for mu in module.uses]):
                used_modules = []
                for aum in additional_used_modules:
                    if aum.name not in [m.name for m in module.uses]:
                        module.uses.append(aum)
                    if aum.uses:
                        used_modules.extend(um.uses)

                additional_used_modules = set(used_modules + module.uses)

        # Now we have just to loop through the list to see which modules use mine
        for module in self.modules:
            for mu in module.uses:
                if module not in mu.used_by:
                    mu.used_by.append(module)

        # Add public var of used modules to availables var
        for module in self.modules:
            for used_module in module.uses:
                if len(used_module.variables):
                    for var in used_module.variables:
                        module.indirect_accessible_var.append(var)

    def expand_derived_types(self):
        for derived_type in self.derived_types:
            if derived_type.extended_structure is not None:
                # Save in the mother structure it extension
                try:
                    struct_to_extend = [s for s in self.derived_types if s.name == derived_type.extended_structure][0]
                except IndexError:
                    print("here")
                struct_to_extend.extension = derived_type
                # save in the extension the mother structure
                derived_type.extension = struct_to_extend

    def generate_namelist_precisions(self, filename="namelist_precisions", exclude_double_precision=True):
        rpe_vars = [var for var in self.variables if var.type == "rpe_var"]
        assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
        assigned_rpe_vars = sorted(assigned_rpe_vars, key=lambda x: x.id)
        line_text = "emulator_variable_precisions(%d) = %d\t! Variable:%24s\tRoutine:%22s\tModule:%16s\n"
        with open(filename, "w") as namelist:
            namelist.write("! namelist variable precisions\n&precisions\n")
            for var in assigned_rpe_vars:
                if var.precision == 52 and exclude_double_precision:
                    continue
                if var.precision is None:
                    continue
                try:
                    if var.untouchable:
                        continue
                except AttributeError:
                    pass
                namelist.write(line_text % (var.id, var.precision, var.name, var.routine.name, var.routine.module.name))
            namelist.write("/\n")

    def assign_ids_to_immutable_variables(self):
        # Loop over rpe-variables and assign an id to those that don't have any, setting them to mutable=False
        last_id = max([v.id for v in self.variables if v.mutable is True])

        counter = last_id + 1
        for variable in self.variables:
            if variable.type == "rpe_var":
                if variable.id is None:
                    variable.id = counter
                    variable.mutable = False
                    counter += 1
