import signal


class GracefulKiller:
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        self.filename = None

    def exit_gracefully(self, signum, frame):
        print("Exit has been activated, will write the checkpoint as soon as possible and exit.")
        self.kill_now = True


class PrecisionAnalysis:
    def __init__(self, remote_hostname=None, remote_user=None, remote_scratch=None, job_template=None, vault=None,
                 reduced_precision_level=10, max_running_jobs=10, experiment_name="PrecisionAnalysis",
                 dummy=False, checkpoint_name=None, assertion_template=None,
                 limit_file=None, analysis_variables=None, forced_variables=None, local_folder=None):
        if dummy is False:
            from AutoRPE.UtilsWorkflow.Communicator import SSH

            self.experiment_name = experiment_name
            self.remote_hostname = remote_hostname
            self.remote_user = remote_user
            self.remote_scratch = remote_scratch
            self.communicator = SSH(user=self.remote_user, host=self.remote_hostname, remote_scratch=self.remote_scratch)
            self.job_template = job_template
            self.vault = vault
            self.reduced_precision_level = reduced_precision_level
            self.max_running_jobs = max_running_jobs
            self.checkpoint_name = checkpoint_name
            self.assertion_template = assertion_template
            self.analysis_variables = analysis_variables
            self.local_folder = local_folder
            if forced_variables is not None:
                self.forced_variables = forced_variables[:]
            else:
                self.forced_variables = None
            if limit_file is not None:
                import pickle as pkl
                self.limit_file = pkl.load(open(limit_file, "rb"))
            else:
                self.limit_file = None

    def run(self):
        from AutoRPE.UtilsWorkflow.WorkflowManagment import BinarySearchJob, ExperimentDriver
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint, recover_checkpoint
        from AutoRPE.UtilsRPE.Functions import store_variable_list
        from AutoRPE.EvaluationScripts.result_explorer import create_explorer
        import time

        vault = self.vault
        # Keeping the variables that have an id (it's precision can be changed in the code)
        variables = [v for v in vault.variables if v.id is not None]

        if self.analysis_variables:
            variables = [v for v in variables if v.id in self.analysis_variables]

        print("Starting analysis with %i variables" % len(variables))
        # Original set including all the variables
        original_set = [v.id for v in variables]

        if self.checkpoint_name is None:
            # Creating the first set that will include all the variables
            starting_job = BinarySearchJob(analysis_set=original_set, forced_variables=self.forced_variables,
                                           reduced_precision_level=self.reduced_precision_level,
                                           communicator=self.communicator, vault=vault, template=self.job_template,
                                           assertion_template=self.assertion_template, limit_file=self.limit_file,
                                           local_folder=self.local_folder)
            # Defining it's level to 0
            starting_job.level = 0
            # Initializing an empty queue for the execution
            experiment = ExperimentDriver(starting_job)
        else:
            # Loading from checkpoint
            print("Resuming from %s" % self.checkpoint_name)
            experiment, starting_job, loops = recover_checkpoint(communicator=self.communicator, vault=vault,
                                                                 filename=self.checkpoint_name)
            # As the checkpoint can come from a different local platform, need to adjust template and local folder.
            for job in experiment.all():
                job.template = self.job_template
                job.local_folder = self.local_folder
            starting_job.vault = vault
            if self.assertion_template:
                for job in experiment.all():
                    if job.assertion_job is not None:
                        job.assertion_job.communicator = self.communicator

        # Make a double-precision run to be sure that the accuracy test works as expected
        check_job = BinarySearchJob(
            analysis_set=[],
            forced_variables=self.forced_variables,
            reduced_precision_level=self.reduced_precision_level,
            communicator=self.communicator,
            vault=vault,
            template=self.job_template,
            assertion_template=self.assertion_template,
            limit_file=self.limit_file,
            local_folder=self.local_folder
        )
        # Submit job and wait until it finishes
        check_job.submit()
        while check_job.remote_status != "COMPLETED" and check_job.remote_status != "FAILED":
            time.sleep(5)

        # Check that the job ran successfully
        if check_job.evaluate_success():
            print("Basic test was successful, continuing with the analysis")
        else:
            raise AssertionError("Basic simulation to check that the accuracy test works properly didn't pass")

        # Define a condition to keep iterating with the refinement
        def condition(_job):
            if _job.status == "FAILED":
                raise AssertionError("Whole set job ended with a failed result. Should check what happened.")
            return _job.status != "SUCCESS"

        # Start dynamic workflow

        try:
            change_counter = 0
            wait = False
            waiting_time = 10
            # Use a signal catcher to handle experiment pause
            killer = GracefulKiller()
            while condition(experiment.seed_job):
                # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
                loop_start_time = time.time()
                # Main loop
                # Sorting the queue to launch first the lower levels (bigger groups).
                experiment.sort(lower_first=True)
                # Launch one step (will loop through the jobs in the queue and take actions according to their status)
                step_change = experiment.step()
                if step_change:
                    experiment.print_status()
                change_counter += step_change
                if change_counter > 10:
                    experiment.checkpoint()
                    change_counter = 0
                loop_end_time = time.time()
                if wait:
                    while loop_end_time - loop_start_time < waiting_time:
                        time.sleep(1)
                        loop_end_time = time.time()
                # If there was a signal during this loop:
                if killer.kill_now:
                    experiment.checkpoint(filename="paused_checkpoint.pkl")
                    print("Exiting...")
                    exit(0)
        except AssertionError as err:
            print(err)
            # For now skip the explorer part
            # Create HTML explorer to make result exploration easier
            # create_explorer(experiment=experiment, vault=vault)
            return False

        # Save lists of variables
        store_variable_list(starting_job.variable_set, "%s_final_set.txt" % self.experiment_name, vault=self.vault)
        store_variable_list(starting_job.banned_variables, "%s_banned_variables.txt" % self.experiment_name, vault=self.vault)

        create_explorer(experiment=experiment, vault=vault)
        # Saving a last checkpoint
        checkpoint(experiment, 1000, filename="end_checkpoint.pkl")
        # Return success
        return True
